package database_client

//go:generate counterfeiter . DatabaseClient

import (
	"database/sql"
	"strings"
)

type DbClient struct {
	connection *sql.DB
}

type DatabaseClient interface {
	ShowGlobalStatus() (map[string]string, error)
	ShowGlobalVariables() (map[string]string, error)
	ShowSlaveStatus() (map[string]string, error)
	SecondsSinceLeaderHeartbeat() (map[string]string, error)
}

func NewDatabaseClient(connection *sql.DB) *DbClient {
	return &DbClient{
		connection: connection,
	}
}

func (dc *DbClient) ShowGlobalStatus() (map[string]string, error) {
	return dc.runKeyValueQuery("SHOW GLOBAL STATUS")
}

func (dc *DbClient) ShowGlobalVariables() (map[string]string, error) {
	return dc.runKeyValueQuery("SHOW GLOBAL VARIABLES")
}

func (dc *DbClient) ShowSlaveStatus() (map[string]string, error) {
	return dc.runSingleRowQuery("SHOW SLAVE STATUS")
}

func (dc *DbClient) SecondsSinceLeaderHeartbeat() (map[string]string, error) {
	return dc.runSingleRowQuery("SELECT NOW() - timestamp AS seconds_since_leader_heartbeat FROM replication_monitoring.heartbeat")
}

func (dc *DbClient) runSingleRowQuery(query string) (map[string]string, error) {
	rows, err := dc.connection.Query(query)
	defer rows.Close()
	if err != nil {
		return nil, err
	}

	result := make(map[string]string)

	columns, err := rows.Columns()
	if err != nil {
		return nil, err
	}

	if rows.Next() {
		columnValues := make([]interface{}, len(columns))
		columnPointers := make([]interface{}, len(columns))
		for i := range columns {
			columnPointers[i] = &columnValues[i]
		}
		err := rows.Scan(columnPointers...)
		if err != nil {
			panic(err)
		}
		for index, col := range columns {
			value := columnValues[index]
			if value == nil {
				value = []uint8("NULL")
			}
			result[strings.ToLower(col)] = string(value.([]uint8))
		}
	}

	return result, nil
}

func (dc *DbClient) runKeyValueQuery(query string) (map[string]string, error) {
	rows, err := dc.connection.Query(query)
	defer rows.Close()
	if err != nil {
		return nil, err
	}

	globalStatusVariables := make(map[string]string)
	for rows.Next() {
		var key, value string
		rows.Scan(&key, &value)
		globalStatusVariables[strings.ToLower(key)] = value

	}

	return globalStatusVariables, nil
}
