package replication_heartbeat

import (
	"database/sql"
	"errors"
	"fmt"
	"strings"

	cfg "github.com/pivotal-cf-experimental/mysql-metrics/config"
	. "github.com/pivotal-cf-experimental/mysql-metrics/logger"
)

const replicationDatabaseName = "replication_monitoring"
const monitoringTableName = replicationDatabaseName + ".heartbeat"
const eventName = replicationDatabaseName + ".replication_monitoring_event"

type MysqlNodeClient struct {
	client *sql.DB
	logger MetricsLogger
	config *cfg.Config
}

func NewMysqlNodeClient(client *sql.DB, logger MetricsLogger, config *cfg.Config) *MysqlNodeClient {
	return &MysqlNodeClient{
		client: client,
		logger: logger,
		config: config,
	}
}

func (mnc *MysqlNodeClient) isLeader() (bool, error) {
	row, err := mnc.client.Query("SHOW SLAVE STATUS")
	if err != nil {

		mnc.logger.Error("show_slave_status.failed", err)
		return false, err
	}
	if row.Next() {
		return false, nil
	}
	return true, nil
}

func (mnc *MysqlNodeClient) InitializeHeartbeat() (bool, error) {
	isLeader, _ := mnc.isLeader()

	if isLeader {
		if !mnc.checkReplicationDatabaseExists() {
			return false, errors.New("check database failed")
		}
		if !mnc.dropTable() {
			return false, errors.New("drop table failed")
		}
		if !mnc.initializeTable() {
			return false, errors.New("create table failed")
		}
		if !mnc.createEvent() {
			return false, errors.New("create event failed")
		}
		return true, nil
	} else {
		mnc.logger.Info("followerNode.heartbeatInitializeSkipped")
		return false, nil
	}
}

func (mnc *MysqlNodeClient) executeQuery(query string, eventType string) bool {
	_, err := mnc.client.Query(query)
	if err != nil {
		mnc.logger.Error(fmt.Sprintf("%s.failed", eventType), err)
		return false
	}
	mnc.logger.Info(fmt.Sprintf("%s.succeeded", eventType))
	return true
}

func (mnc *MysqlNodeClient) checkReplicationDatabaseExists() bool {
	return mnc.executeQuery(fmt.Sprintf("SHOW TABLES IN %s", replicationDatabaseName), "showTables")
}

func (mnc *MysqlNodeClient) dropTable() bool {
	return mnc.executeQuery(fmt.Sprintf("DROP TABLE IF EXISTS %s", monitoringTableName), "dropTableIfExists")
}

func (mnc *MysqlNodeClient) initializeTable() bool {
	success := mnc.executeQuery(fmt.Sprintf(""+
		"CREATE TABLE %s ("+
		"timestamp DATETIME NOT NULL, "+
		"server_id varchar(255))", monitoringTableName), "createTable")
	if !success {
		return success
	}

	return mnc.executeQuery(
		fmt.Sprintf("INSERT INTO %s VALUE (now(), %q)", monitoringTableName, mnc.config.InstanceId),
		"seedTable",
	)
}

func (mnc *MysqlNodeClient) createEvent() bool {
	row, err := mnc.client.Query("SELECT USER()")
	if err != nil {
		mnc.logger.Error("notAbleToAccessDB", err)
		return false
	}
	var userName string
	if row.Next() {
		row.Scan(&userName)
	}

	success := mnc.executeQuery(fmt.Sprintf("DROP EVENT IF EXISTS %s", eventName), "dropEventIfExists")
	if !success {
		return success
	}

	userAndHost := strings.Split(userName, "@")
	return mnc.executeQuery(fmt.Sprintf(
		"CREATE DEFINER = '%s'@'%s' "+
			"EVENT %s "+
			"ON SCHEDULE EVERY 5 SECOND "+
			"ENABLE "+
			"DO UPDATE %s SET timestamp = now()",
		userAndHost[0], "%", eventName, monitoringTableName), "createEvent")
}
