package replication_heartbeat_test

import (
	"database/sql"
	"errors"

	"code.cloudfoundry.org/lager"
	"code.cloudfoundry.org/lager/lagertest"

	"bytes"
	"github.com/DATA-DOG/go-sqlmock"
	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
	cfg "github.com/pivotal-cf-experimental/mysql-metrics/config"
	"github.com/pivotal-cf-experimental/mysql-metrics/logger/loggerfakes"
	"github.com/pivotal-cf-experimental/mysql-metrics/replication_heartbeat"
	"regexp"
)

var _ = Describe("MysqlNodeClient", func() {
	Describe("InitializeHeartbeat", func() {
		var (
			conn             *sql.DB
			mock             sqlmock.Sqlmock
			config           *cfg.Config
			configInstanceId string
			mysqlNodeClient  *replication_heartbeat.MysqlNodeClient
			err              error
			logger           *lagertest.TestLogger
			loggerSourceName string
			mysqlUserAndHost string
			fakeLogger       *loggerfakes.FakeMetricsLogger
		)

		BeforeEach(func() {
			loggerSourceName = "testingLogger"
			conn, mock, err = sqlmock.New()
			Expect(err).NotTo(HaveOccurred())

			logger = lagertest.NewTestLogger(loggerSourceName)
			logger.RegisterSink(lager.NewWriterSink(GinkgoWriter, lager.DEBUG))

			mysqlUserAndHost = "mysql-metrics@localhost"

			fakeLogger = new(loggerfakes.FakeMetricsLogger)
			configInstanceId = "my-instance-id"
			config = &cfg.Config{
				InstanceId: configInstanceId,
			}

			mysqlNodeClient = replication_heartbeat.NewMysqlNodeClient(conn, fakeLogger, config)
		})

		Context("If current node is leader node", func() {
			BeforeEach(func() {
				mock.ExpectQuery(`SHOW SLAVE STATUS`).WillReturnRows(sqlmock.NewRows([]string{}))
			})

			Context("checks replication_database", func() {
				It("logs error and fails if does not exists", func() {
					mock.ExpectQuery(`SHOW TABLES IN replication_monitoring`).WillReturnError(errors.New("Unknown database 'replication_monitoring'"))

					mysqlNodeClient.InitializeHeartbeat()

					expectLogsToContainSubstring(fakeLogger, "showTables.failed")
				})

				It("logs a success message if exists", func() {
					row := sqlmock.NewRows([]string{"Tables_in_replication_monitoring"}).
						AddRow("monitoring")
					mock.ExpectQuery(`SHOW TABLES IN replication_monitoring`).WillReturnRows(row)

					mysqlNodeClient.InitializeHeartbeat()

					expectLogsToContainSubstring(fakeLogger, "showTables.succeeded")
				})
			})

			Context("replication_database exists", func() {
				BeforeEach(func() {
					row := sqlmock.NewRows([]string{"Tables_in_replication_monitoring"}).
						AddRow("monitoring")
					mock.ExpectQuery(`SHOW TABLES IN replication_monitoring`).WillReturnRows(row)
				})

				Context("drop existing monitoring table", func() {
					It("drop existing monitoring table named `heartbeat` succeeded", func() {
						mock.ExpectQuery(`DROP TABLE IF EXISTS replication_monitoring.heartbeat`).WillReturnRows(sqlmock.NewRows([]string{}))
						mysqlNodeClient.InitializeHeartbeat()

						expectLogsToContainSubstring(fakeLogger, "dropTableIfExists.succeeded")
					})
					It("drop existing monitoring table named `heartbeat` failed", func() {
						mock.ExpectQuery(`SHOW TABLES IN replication_monitoring`).WillReturnError(errors.New("DROP command denied to user 'mysql-metrics'@'localhost' for table 'sometablename'"))

						mysqlNodeClient.InitializeHeartbeat()

						expectLogsToContainSubstring(fakeLogger, "dropTableIfExists.failed")
					})
				})

				Context("creates monitoring table", func() {
					BeforeEach(func() {
						mock.ExpectQuery(`DROP TABLE IF EXISTS replication_monitoring.heartbeat`).WillReturnRows(sqlmock.NewRows([]string{}))
					})

					It("succeeded in creating the `heartbeat` table", func() {
						mock.ExpectQuery(`CREATE TABLE replication_monitoring.heartbeat`).WillReturnRows(sqlmock.NewRows([]string{}))
						mock.ExpectQuery(
							regexp.QuoteMeta(`INSERT INTO replication_monitoring.heartbeat VALUE (now(), "my-instance-id")`)).
							WillReturnRows(sqlmock.NewRows([]string{}))

						mysqlNodeClient.InitializeHeartbeat()

						Expect(mock.ExpectationsWereMet()).NotTo(HaveOccurred())
						mock.ExpectationsWereMet()
						expectLogsToContainSubstring(fakeLogger, "createTable.succeeded")
						expectLogsToContainSubstring(fakeLogger, "seedTable.succeeded")
					})

					It("failed in creating the `heartbeat` table", func() {
						mock.ExpectQuery(`CREATE TABLE replication_monitoring.heartbeat`).WillReturnError(errors.New("CREATE command denied to user"))

						mysqlNodeClient.InitializeHeartbeat()

						expectLogsToContainSubstring(fakeLogger, "createTable.failed")
						mock.ExpectationsWereMet()
					})

					It("failed in seeding the `heartbeat` table", func() {
						mock.ExpectQuery(`CREATE TABLE replication_monitoring.heartbeat`).WillReturnRows(sqlmock.NewRows([]string{}))
						mock.ExpectQuery(`INSERT INTO replication_monitoring.heartbeat`).WillReturnError(errors.New("INSERT went poop"))

						mysqlNodeClient.InitializeHeartbeat()

						expectLogsToContainSubstring(fakeLogger, "seedTable.failed")
					})
				})

				Context("creates replication_monitoring_event", func() {
					BeforeEach(func() {
						mock.ExpectQuery(`DROP TABLE IF EXISTS replication_monitoring.heartbeat`).WillReturnRows(sqlmock.NewRows([]string{}))
						mock.ExpectQuery(`CREATE TABLE replication_monitoring.heartbeat`).WillReturnRows(sqlmock.NewRows([]string{}))
						mock.ExpectQuery(`INSERT INTO replication_monitoring.heartbeat`).WillReturnRows(sqlmock.NewRows([]string{}))

						row := sqlmock.NewRows([]string{"user()"}).AddRow(mysqlUserAndHost)
						mock.ExpectQuery(`SELECT USER()`).WillReturnRows(row)
					})

					It("succeeded in creating event", func() {
						mock.ExpectQuery(`DROP EVENT IF EXISTS replication_monitoring.replication_monitoring_event`).WillReturnRows(sqlmock.NewRows([]string{}))
						mock.ExpectQuery(`CREATE DEFINER`).WillReturnRows(sqlmock.NewRows([]string{}))

						mysqlNodeClient.InitializeHeartbeat()

						expectLogsToContainSubstring(fakeLogger, "createEvent.succeeded")
					})

					It("failed in creating event", func() {
						mock.ExpectQuery(`DROP EVENT IF EXISTS replication_monitoring.replication_monitoring_event`).WillReturnRows(sqlmock.NewRows([]string{}))
						mock.ExpectQuery(`CREATE DEFINER`).WillReturnError(errors.New("CREATE command denied to user"))

						mysqlNodeClient.InitializeHeartbeat()

						expectLogsToContainSubstring(fakeLogger, "createEvent.failed")
					})
				})
			})
		})

		Context("If current node is not a leader node", func() {
			BeforeEach(func() {
				mock.ExpectQuery(`SHOW SLAVE STATUS`).WillReturnRows(sqlmock.NewRows([]string{"some-slave-status-variable"}).AddRow("iam_groot"))
			})

			It("should report that it failed to initialize heartbeat", func() {
				mysqlNodeClient.InitializeHeartbeat()

				expectLogsToContainSubstring(fakeLogger, "followerNode.heartbeatInitializeSkipped")
			})
		})
	})
})

func expectLogsToContainSubstring(fakeLogger *loggerfakes.FakeMetricsLogger, expected string) {
	errorCount := fakeLogger.ErrorCallCount()
	infoCount := fakeLogger.InfoCallCount()

	Expect(errorCount + infoCount).ToNot(Equal(0))
	var buffer bytes.Buffer

	for i := 0; i < errorCount; i++ {
		log, _ := fakeLogger.ErrorArgsForCall(i)
		buffer.WriteString(log)
		buffer.WriteString("\n")
	}

	for i := 0; i < infoCount; i++ {
		log := fakeLogger.InfoArgsForCall(i)
		buffer.WriteString(log)
		buffer.WriteString("\n")
	}

	Expect(buffer.String()).To(ContainSubstring(expected))
}
